from collections import namedtuple
import platform
import json
import encodings
import hashlib
import socket
import logging

from toolz.curried import (
    compose, pipe, map, first, concatv,
)

from .common import (
    Null,
)

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

_py_attr = [
    'branch', 'build', 'compiler', 'implementation', 'revision', 'version'
]
_sig_attr = pipe(
    concatv(
        ['hostname', 'platform', 'architecture', 'machine', 'processor'],
        pipe(_py_attr, map(lambda a: f'python_{a}'), tuple),
    )
)
Signature = namedtuple('Signature', _sig_attr)

def host_signature():
    return Signature(
        socket.gethostname(),
        platform.platform(),
        platform.architecture(),
        platform.machine(),
        platform.processor(),
        *pipe(
            _py_attr,
            map(lambda a: f'python_{a}'),
            map(lambda a: getattr(platform, a)),
            map(lambda f: f()),
            tuple,
        ),
    )

def verify_signature_dict(sig):
    if not type(sig) is dict:
        log.error('Signature provided is not a dictionary')
        return False

    missing_keys = set(_sig_attr) - set(sig)
    if missing_keys:
        log.error(
            f'Signature missing keys: '
            f'{", ".join(sorted(missing_keys))}'
        )
        return False
    return True

def signature_from_dict(sig_dict):
    if not verify_signature_dict(sig_dict):
        return None
    return Signature(**sig_dict)
from_dict = signature_from_dict

def maybe_dict(t):
    try:
        return dict(t)
    except TypeError:
        return Null

def signature_from_tuple(sig_tuple):
    return pipe(
        sig_tuple,
        maybe_dict,
        signature_from_dict,
    )
from_tuple = signature_from_tuple

def signature_to_dict(sig):
    return sig._asdict()
to_dict = signature_to_dict

def signature_to_tuple(sig):
    return tuple(sig._asdict())
to_tuple = signature_to_tuple

utf8 = compose(first, encodings.utf_8.encode)

def signature_to_hash(sig):
    return pipe(
        sig,
        to_tuple,
        json.dumps,
        utf8,
        hashlib.sha512,
        lambda h: h.hexdigest(),
    )
to_hash = signature_to_hash



import concurrent.futures

from toolz.curried import (
    curry, concatv,
)
import gevent
import gevent.pool

from .common import vcall

@curry
def thread_map(func, iterable, *iterables, **tpe_kw):
    with concurrent.futures.ThreadPoolExecutor(**tpe_kw) as executor:
        for value in executor.map(func, *concatv((iterable,), *iterables)):
            yield value

@curry
def process_map(func, iterable, *iterables, **tpe_kw):
    with concurrent.futures.ProcessPoolExecutor(**tpe_kw) as executor:
        for value in executor.map(func, *concatv((iterable,), *iterables)):
            yield value

@curry
def thread_vmap(func, iterable, *iterables, **tpe_kw):
    yield from thread_map(vcall(func), iterable, *iterables, **tpe_kw)

@curry
def gevent_map(func, iterable, *iterables, max_workers=10):
    pool = gevent.pool.Pool(max_workers)
    return pool.map(func, *concatv((iterable,), *iterables))

@curry
def gevent_vmap(func, iterable, *iterables, max_workers=10):
    pool = gevent.pool.Pool(max_workers)
    return pool.map(vcall(func), *concatv((iterable,), *iterables))

def pmap(ptype: str):
    return {
        'thread': thread_map,
        'gevent': gevent_map,
    }.get(ptype, thread_map)



import re

import jinja2
import markdown
from toolz.curried import (
    pipe, complement, juxt, merge,
    interleave, partition, reduce, map,
)

from .. import yaml
from ..common import vmap, vcall, maybe_pipe, short_circuit

def monotonic(seq, cmp=lambda a, b: a <= b):
    return all(cmp(seq[i], seq[i + 1]) for i in range(0, len(seq) - 1))

class YamlDataError(ValueError):
    pass

start_re = re.compile(r'--yaml--', re.IGNORECASE)
end_re = re.compile(r'--endyaml--', re.IGNORECASE)

def not_enough_tags(lines):
    s, e = start_re.pattern, end_re.pattern
    return (
        lines.count(s) != lines.count(e),
        f'Start and end YAML tags do not agree: {s} start != {e} end'
    )

def start_indices(lines):
    return [i for i, l in enumerate(lines) if start_re.search(l)]

def end_indices(lines):
    return [i for i, l in enumerate(lines) if end_re.search(l)]

def all_indices(lines):
    return list(interleave(
        (start_indices(lines), end_indices(lines))
    ))

def not_correct_order(lines):
    valid = pipe(
        all_indices(lines),
        complement(monotonic),
    )
    return (
        valid, 'Start and end tags are not in the correct order.'
    )

def check_lines(lines):
    errors = [
        msg for boolean, msg in juxt(
            not_correct_order, not_enough_tags,
        )(lines) if boolean
    ]

    if errors:
        return False, '\n'.join(errors)
    return True, ''

def yaml_data(lines):
    def render(raw, data):
        return jinja2.Template(raw).render(**data)

    from toolz.curried import do
    
    return maybe_pipe(
        all_indices(lines),
        short_circuit(bool),  # catch null YAML early
        partition(2),
        vmap(lambda s, e: lines[s + 1:e]),
        map('\n'.join),
        map(yaml.load),
        merge,
        # lambda lines: '\n'.join(lines),
        # lambda raw: (raw, yaml.load(raw)),
        # vcall(render),
        # lambda raw: (raw, yaml.load(raw)),
        # vcall(render),
        # lambda data: data[1],
    ) or {}

def non_yaml_lines(lines):
    return pipe(
        [-1] + all_indices(lines) + [len(lines)],
        partition(2),
        vmap(lambda s, e: lines[s + 1:e]),
        reduce(lambda a, b: a + b)
    )

class YamlDataPreprocessor(markdown.preprocessors.Preprocessor):
    '''Parse out the YAML metadata blocks from the markdown

    A YAML metadata block:
    
    - Is delimited by:
        - a line `--yaml--` at the start
        - and a line `--endyaml--` at the end
    
    '''
    def run(self, lines):
        '''Parse YAML metadata blocks and store in self.markdown.metadata'''
        valid, msg = check_lines(lines)
        if not valid:
            raise YamlDataError(msg)

        meta = yaml_data(lines)
        if self.md.meta is not None:
            self.md.meta = merge(self.md.meta or {}, meta)
        else:
            self.md.meta = meta

        return non_yaml_lines(lines)


class YamlDataExtension(markdown.Extension):
    def extendMarkdown(self, md, md_globals):
        if not hasattr(md, 'meta'):
            md.meta = None
        md.preprocessors.register(
            YamlDataPreprocessor(md), "yaml_data", 90
        )

def makeExtension(*args, **kwargs):
    return YamlDataExtension(*args, **kwargs)



import coloredlogs
from toolz import merge

def setup_logging(loglevel: str, **config_kw):
    fmt = (
        '{asctime} {levelname: <6} [{name}:{lineno: >4}]  {message}'
    )
    datefmt = '%Y-%m-%d %H:%M:%S'
    kw = merge({
        'level': loglevel.upper(),
        'datefmt': datefmt,
        'fmt': fmt,
        'style': '{',
    }, config_kw)
    coloredlogs.install(**kw)




'''Tools for dealing with IP data

'''
from ipaddress import ip_interface

import click
from toolz.curried import (
    pipe, compose, map, filter,
)

from ..common import (
    is_ip, strip_comments, strip, sortips,
    get_ips_from_file, get_ips_from_str, do_nothing,
    clipboard_copy, zpad, unzpad,
)
from .common import (
    get_input_content
)
from ..yaml import dump as dump_yaml

cb_copy_ensure_nl = compose(
    clipboard_copy,
    lambda c: c if c.endswith('\n') else c + '\n'
)

@click.command()
@click.argument(
    'patha',
    type=click.Path(exists=True),
)
@click.argument(
    'pathb',
    type=click.Path(exists=True),
)
def diff_ips(patha, pathb):
    '''Given PATHA and PATHB of IPs, print difference (A-B)

    '''
    ips_a = set(get_ips_from_file(patha))
    ips_b = set(get_ips_from_file(pathb))

    pipe(
        ips_a - ips_b,
        sortips,
        '\n'.join,
        print,
    )

@click.command(
    help=('Given PATHA and PATHB of IPs, print intersection (A & B)'),
)
@click.argument(
    'patha',
    type=click.Path(exists=True),
)
@click.argument(
    'pathb',
    type=click.Path(exists=True),
)
def int_ips(patha, pathb):
    ips_a = set(get_ips_from_file(patha))
    ips_b = set(get_ips_from_file(pathb))

    pipe(
        ips_a & ips_b,
        sortips,
        '\n'.join,
        print,
    )

@click.command()
@click.argument(
    'inpath',
    required=False,
    type=click.Path(exists=True),
)
@click.option(
    '-K', '--keepcomments', is_flag=True,
    help=('Keep the in-line comments'),
)
@click.option(
    '-C', '--clipboard', is_flag=True,
    help=('Get IPs from clipboard, send sorted to clipboard'),
)
def sort_ips(inpath, keepcomments, clipboard):
    '''Given a list of IPs (one per line) from a file path (INPATH), the
    clipboard (-C), or stdin if nothing provided, print in sorted
    order (to stdout unless -C is provided).

    '''
    content = get_input_content(inpath, clipboard)

    return pipe(
        content.splitlines(),
        map(do_nothing if keepcomments else strip_comments),
        filter(lambda l: l.strip()),
        filter(compose(is_ip, strip, strip_comments)),
        sortips,
        '\n'.join,
        print if not clipboard else cb_copy_ensure_nl,
    )

@click.command()
@click.argument(
    'inpath',
    required=False,
    type=click.Path(exists=True),
)
@click.option(
    '-s', '--slash', type=int, default=24,
    help='Find the networks in the list of IPs of the given size (16, 24)'
)
@click.option(
    '-c', '--from-clipboard', is_flag=True,
    help=('Get IPs from clipboard'),
)
@click.option(
    '-C', '--to-clipboard', is_flag=True,
    help=('Send sorted to clipboard'),
)
@click.option(
    '--yaml', is_flag=True,
    help='Output should be in yaml form'
)
def get_subnets(inpath, slash, from_clipboard, to_clipboard, yaml):
    ''''Given a list of IP addresses from a file path (INPATH), the
    clipboard (-C), or stdin (if nothing provided), print in sorted
    order (to stdout unless -C is provided) all the IP networks of a
    certain size (-s {16, 24})

    '''
    content = get_input_content(inpath, from_clipboard)

    def get_network(ip):
        return ip_interface(ip + f'/{slash}').network
            
    return pipe(
        get_ips_from_str(content),
        map(get_network),
        set,
        sorted,
        map(str),
        compose(dump_yaml, list) if yaml else '\n'.join,
        print if not to_clipboard else cb_copy_ensure_nl,
    )


@click.command()
@click.argument(
    'inpath',
    required=False,
    type=click.Path(exists=True),
)
@click.option(
    '-c', '--from-clipboard', is_flag=True,
    help=('Get IPs from clipboard'),
)
@click.option(
    '-C', '--to-clipboard', is_flag=True,
    help=('Send sorted to clipboard'),
)
@click.option(
    '--stdout', is_flag=True,
    help='Force output to stdout',
)
@click.option(
    '--no-sort', is_flag=True,
    help='Keep IPs unsorted',
)
@click.option(
    '-u', '--unique', is_flag=True,
    help='Print only unique IPs',
)
def get_ips(inpath, from_clipboard, to_clipboard, stdout, no_sort, unique):
    '''Given a text block containing IPs from a file path (INPATH), the
    clipboard (-c), or stdin if nothing provided, print in sorted
    order (to stdout unless -C is provided) all the IPs

    '''
    content = get_input_content(inpath, from_clipboard)
            
    return pipe(
        get_ips_from_str(content),
        set if unique else do_nothing,
        do_nothing if no_sort else sortips,
        '\n'.join,
        print if stdout or not to_clipboard else cb_copy_ensure_nl,
    )


@click.command()
@click.argument(
    'inpath',
    required=False,
    type=click.Path(exists=True),
)
@click.option(
    '-c', '--from-clipboard', is_flag=True,
    help=('Get IPs from clipboard'),
)
@click.option(
    '-C', '--to-clipboard', is_flag=True,
    help=('Send sorted to clipboard'),
)
@click.option(
    '--stdout', is_flag=True,
    help='Force output to stdout',
)
@click.option(
    '--no-sort', is_flag=True,
    help='Keep IPs unsorted',
)
@click.option(
    '-u', '--unique', is_flag=True,
    help='Print only unique IPs',
)
def zpad_ips(inpath, from_clipboard, to_clipboard, stdout, no_sort, unique):
    '''Given a text block containing IPs from a file path (INPATH), the
    clipboard (-c), or stdin if nothing provided, print in sorted
    order (to stdout unless -C is provided) all the IPs with octets
    padded with zeros.

    '''
    content = get_input_content(inpath, from_clipboard)
            
    return pipe(
        get_ips_from_str(content),
        set if unique else do_nothing,
        do_nothing if no_sort else sortips,
        map(zpad),
        '\n'.join,
        print if stdout or not to_clipboard else cb_copy_ensure_nl,
    )

@click.command()
@click.argument(
    'inpath',
    required=False,
    type=click.Path(exists=True),
)
@click.option(
    '-c', '--from-clipboard', is_flag=True,
    help=('Get IPs from clipboard'),
)
@click.option(
    '-C', '--to-clipboard', is_flag=True,
    help=('Send sorted to clipboard'),
)
@click.option(
    '--stdout', is_flag=True,
    help='Force output to stdout',
)
@click.option(
    '--no-sort', is_flag=True,
    help='Keep IPs unsorted',
)
@click.option(
    '-u', '--unique', is_flag=True,
    help='Print only unique IPs',
)
def unzpad_ips(inpath, from_clipboard, to_clipboard, stdout, no_sort, unique):
    '''Given a text block containing IPs from a file path (INPATH), the
    clipboard (-c), or stdin if nothing provided, print in sorted
    order (to stdout unless -C is provided) all the zero-padded IPs
    formatted as normal IPs.

    '''
    content = get_input_content(inpath, from_clipboard)
            
    return pipe(
        get_ips_from_str(content),
        set if unique else do_nothing,
        do_nothing if no_sort else sortips,
        map(unzpad),
        '\n'.join,
        print if stdout or not to_clipboard else cb_copy_ensure_nl,
    )



import re
import logging

from markdown import Extension
from markdown.preprocessors import Preprocessor
from toolz.curried import (
    pipe, compose, curry, complement, juxt, merge,
    interleave, partition, reduce, map, do, filter,
)

from ..common import (
    vmap, vfilter, vcall, maybe_pipe,
)
from .. import yaml

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

TABLE_RE = re.compile(r'^\s*\|==\[\[(?P<attrib>(.*?))\]\]==\|\s*$')
TH_RE = re.compile(
    r'^\s*(?:\[\[(?P<ra1>.*)\]\])?\s*\|=\s+(?P<th>(?:\S[\S ]*?\s+\|)+\s+\S[\S ]*?)\s+=\|\s*(?:\[\[(?P<ra2>.*)\]\])?\s*$'
)
TR_RE = re.compile(
    r'^\s*(?:\[\[(?P<ra1>.*)\]\])?\s*\|\|\s+(?P<tr>(?:\S[\S ]*?\s+\|)+\s+\S[\S ]*?)\s+\|\|\s*(?:\[\[(?P<ra2>.*)\]\])?\s*$'
)

ATTR_RE = re.compile(
    r'\s*(?:\[\[(?P<a>.*)\]\])\s*'
)

def attr_dict(s):
    return dict(yaml.load('{' + s + '}') if (s or '').strip() else {})

def attr_html(d):
    return pipe(
        d.items(),
        vmap(lambda k, v: f'{k}="{v}"'),
        ' '.join,
    )

def parse_tables(lines):
    table_lines = pipe(
        enumerate(lines),
        vmap(lambda i, l: (i, TABLE_RE.match(l))),
        vfilter(lambda i, m: m),
        tuple,
    )

    # log.debug(table_lines)

    ntables = 0

    for start_index, match in table_lines:
        table = {'start_index': start_index}
        log.debug(f'start index: {start_index}')
        attr = attr_dict(match.groupdict()['attrib'])
        table['caption'] = attr.pop('caption', '')
        table['attr'] = attr

        for i, line in enumerate(lines[start_index + 1:], start_index + 1):
            th_match = TH_RE.match(line)
            tr_match = TR_RE.match(line)
            if th_match:
                match_dict = th_match.groupdict()
                table['header'] = {
                    'line': i,
                    'tr': pipe(
                        th_match.groupdict()['th'].split('|'),
                        map(lambda s: s.strip()),
                        map(lambda s: (ATTR_RE.search(s),
                                       ATTR_RE.sub('', s))),
                        vmap(lambda m, s:
                             (merge(
                                 {'scope': 'col'},
                                 attr_dict(m.groupdict()['a']) if m else {},
                             ), s)),
                        vmap(lambda attr, td: {'td': td, 'attr': attr}),
                        tuple,
                    ),
                    'attr': pipe(
                        ['ra1', 'ra2'],
                        map(lambda k: match_dict.get(k, '')),
                        filter(None),
                        map(attr_dict),
                        merge,
                    ),
                }

            elif tr_match:
                match_dict = tr_match.groupdict()
                row = {
                    'line': i,
                    'tr': pipe(
                        match_dict['tr'].split('|'),
                        map(lambda s: s.strip()),
                        map(lambda s: (ATTR_RE.search(s),
                                       ATTR_RE.sub('', s))),
                        vmap(lambda m, s: (
                            attr_dict(m.groupdict()['a']) if m else {}, s
                        )),
                        vmap(lambda attr, td: {'attr': attr, 'td': td}),
                        tuple,
                    ),
                    'attr': pipe(
                        ['ra1', 'ra2'],
                        map(lambda k: match_dict.get(k, '')),
                        filter(None),
                        map(attr_dict),
                        merge,
                    )
                }

                table.setdefault('rows', []).append(row)

            else:
                # This is the case where we have gotten to a non-table
                # line of content, thus we are done with the table
                table['end_index'] = i
                log.debug(f'table end: {i}')
                ntables += 1
                yield table
                break
        
    if ntables < len(table_lines):
        # This is the case where the last line of content is a table
        # line
        log.debug('finishing table')
        table['end_index'] = i + 1
        yield table

@curry
def tr_to_html(data_tag: str, tr: dict):
    return (
        f'<tr {attr_html(tr.get("attr", {}))}>' + pipe(
            tr['tr'],
            map(lambda tr: (
                f'<{data_tag} {attr_html(tr["attr"])} >{tr["td"]}</{data_tag}>'
            )),
            ' '.join,
        ) + '</tr>'
    )

def table_to_html(lines):
    new_lines = lines[:]
    tables = list(parse_tables(lines))
    for table in tables:
        new_lines[table['start_index']] = (
            f'<table {attr_html(table["attr"])} > {table["caption"]}'
        )

        header = table.get('header', {})
        if header:
            new_lines[header['line']] = tr_to_html('th', header)

        for row in table.get('rows', []):
            new_lines[row['line']] = tr_to_html('td', row)

    end_indexes = [t['end_index'] for t in tables]
    for index in end_indexes[::-1]:
        new_lines.insert(index, '</table>')

    return new_lines

def tr_to_pandoc(tr):
    return (
        '| ' + pipe(
            tr['tr'],
            map(lambda tr: tr["td"]),
            ' | '.join,
        ) + ' |'
    )

def table_to_pandoc(lines):
    new_lines = lines[:]
    tables = list(parse_tables(lines))
    header_seps = []
    for table in tables:
        new_lines[table['start_index']] = ''
        header = table.get('header', {})
        if header:
            new_lines[header['line']] = tr_to_pandoc(header)

        if not (header or table['rows']):
            # We have a table start with neither a header or data
            # rows, so just skip and move on
            continue

        header_sep_index = (
            header['line'] if header else table['rows'][0]['line']
        ) + 1
        header_sep_length = (
            len(header['tr']) if header else len(table['rows'][0]['tr'])
        )
        header_seps.append((
            header_sep_index,
            '|' + '|'.join(['---'] * header_sep_length) + '|'
        ))

        for row in table.get('rows', []):
            new_lines[row['line']] = tr_to_pandoc(row)

    for index, sep in header_seps[::-1]:
        new_lines.insert(index, sep)
        
    return new_lines

class TablePreprocessor(Preprocessor):
    def run(self, lines):
        return table_to_html(lines)
            
class SimpleTableExtension(Extension):
    # def extendMarkdown(self, md, md_globals):
    def extendMarkdown(self, md):
        md.preprocessors.register(
            TablePreprocessor(md), "simpletable", 10
        )
        

def makeExtension(*args, **kwargs):
    return SimpleTableExtension(*args, **kwargs)




from typing import Tuple
import logging

import markdown
from toolz.curried import merge

from .. import yaml

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

def meta_and_content(lines: list) -> Tuple[list]:
    '''Split markdown content lines into metadata lines and markdown
    content lines

    Example:

    >>> meta_and_content("""---
    ... a: 1
    ... b: 2
    ... ---
    ... Some **markdown**
    ... """.splitlines())
    (['a: 1', 'b: 2'], ['Some **markdown**'])

    '''
    if lines[0] != '---':
        return [], lines

    stripped = [l.strip() for l in lines]
    
    if '---' not in stripped:
        log.error('Metadata start block with no end block')
        return [], lines

    index = stripped[1:].index('---') + 1
    return lines[1:index], lines[index + 1:]

class MetaYamlPreprocessor(markdown.preprocessors.Preprocessor):
    """Preprocess markdown content with YAML metadata parsing.

    YAML block is delimited by '---' at start and '---' at end.

    """

    def run(self, lines: list) -> list:
        meta_lines, lines = meta_and_content(lines)

        if meta_lines:
            meta = yaml.load('\n'.join(meta_lines))
            if self.md.meta is not None:
                self.md.meta = merge(self.md.meta, meta)
            else:
                self.md.meta = meta
        return lines

class MetaYamlExtension(markdown.Extension):
    """Extension for parsing YAML metadata part with Python-Markdown."""

    def extendMarkdown(self, md: markdown.Markdown, *args, **kwargs):
        if not hasattr(md, 'meta'):
            md.meta = None
        md.preprocessors.register(
            MetaYamlPreprocessor(md), "meta_yaml", 100
        )

def makeExtension(*args, **kwargs):
    return MetaYamlExtension(*args, **kwargs)



'''Tools for dealing with text data
'''
from pathlib import Path

import click
from toolz.curried import pipe

from ..common import difflines, intlines

@click.command(
    help=('Given PATHA and PATHB of text, print difference (A-B)'),
)
@click.argument(
    'patha',
    type=click.Path(exists=True),
)
@click.argument(
    'pathb',
    type=click.Path(exists=True),
)
def diff_lines(patha, pathb):
    pipe(
        difflines(
            Path(patha).expanduser().read_text(),
            Path(pathb).expanduser().read_text(),
        ),
        '\n'.join,
        print,
    )

@click.command(
    help=('Given PATHA and PATHB of text, print intersection (A & B)'),
)
@click.argument(
    'patha',
    type=click.Path(exists=True),
)
@click.argument(
    'pathb',
    type=click.Path(exists=True),
)
def int_lines(patha, pathb):
    pipe(
        intlines(
            Path(patha).expanduser().read_text(),
            Path(pathb).expanduser().read_text(),
        ),
        '\n'.join,
        print,
    )




import urllib
import functools
import logging
import pprint
from typing import Callable, List

import requests
from toolz.curried import (
    pipe, curry, partial, concatv, compose, map, filter, merge
)
from pyrsistent import pmap

from .common import (
    maybe_pipe, vmap, maybe_first, do_nothing, is_dict, vcall, Null,
)

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

class ResponseError(ValueError):
    pass

class TokenAuth(requests.auth.AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, request):
        request.headers['Authorization'] = 'Bearer {}'.format(self.token)
        return request

class Api:
    def __init__(self, base_url, auth, session, *, method_kw=None):
        self.base_url = base_url
        self.auth = auth
        self.session = session
        self.method_kw = method_kw or {}

    def __call__(self, *parts, **kw):
        return Endpoint(
            self, parts, **merge(self.method_kw, kw)
        )

def link_next(response):
    return maybe_pipe(
        requests.utils.parse_header_links(
            response.headers.get('Link', '')
        ),
        filter(lambda d: d.get('rel', '').lower() == 'next'),
        maybe_first,
        lambda d: d['url'],
    )

def method_logger(name, method, *, log_func=None):
    @functools.wraps(method)
    def wrapper(*a, **kw):
        (log_func or log.debug)(f'{name.upper()}: {kw}')
        return method(*a, **kw)
    return wrapper

def namespace_data(data):
    def ns_dict(ns, d):
        return pipe(
            d.items(),
            vmap(lambda k, v: (f'{ns}[{k}]', v)),
            dict,
        )
    if is_dict(data):
        return pipe(
            data.items(),
            vmap(lambda k, v: ns_dict(k, v) if is_dict(v) else {k: v}),
            vcall(merge),
        )
    return data
                    
class Endpoint:
    def __init__(self, api, parts, **kwargs):
        self.api = api
        self.parts = tuple(parts)

        self.kwargs = pipe(
            merge(
                {'url': self.url, 'auth': self.api.auth},
                kwargs,
            ),
            pmap,
        )

        for name in ['get', 'post', 'put', 'delete', 'head',
                     'options', 'patch']:
            setattr(
                self, name, self.method(name, **self.kwargs)
            )
            setattr(
                self, f'maybe_{name}', self.maybe_method(name, **self.kwargs)
            )

    def __call__(self, *parts, **kw):
        return Endpoint(
            self.api, tuple(concatv(self.parts, parts)), **kw
        )

    def method(self, name, **orig_kw):
        def caller(*args, **kw):
            method = getattr(self.api.session, name)
            kw = merge(orig_kw, kw)
            if 'data' in kw:
                kw['data'] = pipe(
                    kw.pop('data'),
                    namespace_data,
                )
            # Preserve JSON data passed in to Api object
            if 'json' in self.api.method_kw:
                kw['json'] = merge(self.api.method_kw['json'],
                                   kw.get('json', {}))
            return method_logger(name, method)(*args, **kw)
        return caller

    def maybe_method(self, name, **orig_kw):
        class Outcome:
            def __init__(self, ok, result, *errors):
                self.ok = ok
                self.result = result
                self.errors = errors

            def __bool__(self):
                return bool(self.ok)

        def caller(*a, **kw):
            try:
                return Outcome(True, self.method(name, **orig_kw)(*a, **kw))
            except OSError as error:
                return Outcome(False, Null, error)
        return caller
        
    @property
    def url(self):
        base = self.api.base_url
        if not base.endswith('/'):
            base += '/'
        return urllib.parse.urljoin(
            base, '/'.join(pipe(self.parts, map(str)))
        )

    @curry
    def iter(self, method, *, url=None,
             iter_f=lambda resp: resp.json(),
             link_next_f=link_next, **requests_kw):
        response = getattr(self, method)(**(
            merge(requests_kw, {'url': url} if url else {})
        ))
        if response.status_code != 200:
            content = response.content.decode()
            raise ResponseError(
                f'Response code error: {response.status_code}\n\n'
                f'{content[:200]}\n'
                '...\n'
                '...\n'
                f'{content[-200:]}'
            )

        for value in iter_f(self, response):
            yield value

        next_url = link_next(response)
        if next_url:
            yield from self.iter(
                method, url=next_url, iter_f=iter_f, **requests_kw
            )

class ResourceEndpoint(Endpoint):
    def __init__(self, endpoint: Endpoint, data: dict, form_key: str):
        self.data = data
        self.form_key = form_key

        super().__init__(endpoint.api, endpoint.parts)
        
def empty_dict(*a, **kw):
    return {}

class IdResourceEndpoint(Endpoint):
    def __init__(self, parent: Endpoint, data: dict, form_key: str,
                 id_key: str,
                 meta_f: Callable[[dict], dict] = empty_dict,
                 unpack_f: Callable[[dict], dict] = do_nothing,
                 single_unpack_f: Callable[[dict], dict] = do_nothing):
        self.parent = parent
        self.data = merge(data, {'metadata': meta_f(data)})
        self.form_key = form_key
        self.id_key = id_key
        self.meta_f = meta_f
        self.unpack_f = unpack_f
        self.single_unpack_f = single_unpack_f

        if id_key not in data:
            log.error(f'ID key {id_key} not in data: {data}')

        super().__init__(parent.api, parent.parts + (data[id_key],))

    @classmethod
    @curry
    def from_multiple_response(cls, parent: Endpoint,
                               response: requests.Response, *,
                               form_key=None, id_key='id',
                               meta_f=empty_dict,
                               unpack_f=do_nothing,
                               single_unpack_f=do_nothing):
        for d in unpack_f(response.json()):
            yield cls(parent, d, form_key, id_key, meta_f=meta_f,
                      unpack_f=unpack_f, single_unpack_f=single_unpack_f)

    @classmethod
    @curry
    def from_single_response(cls, parent: Endpoint,
                             response: requests.Response, *,
                             form_key=None, id_key='id',
                             meta_f=do_nothing,
                             unpack_f=do_nothing,
                             single_unpack_f=do_nothing):
        data = unpack_f(response.json())
        return cls(parent, data, form_key, id_key, meta_f=meta_f,
                   unpack_f=unpack_f, single_unpack_f=single_unpack_f)

    def refresh(self, **get_kw):
        return IdResourceEndpoint(
            self.parent, self.single_unpack_f(self.get(**get_kw).json()),
            self.form_key, self.id_key, meta_f=self.meta_f,
            unpack_f=self.unpack_f, single_unpack_f=self.single_unpack_f,
        )

@curry
def update_endpoint(endpoint: IdResourceEndpoint, update: dict, *,
                    body_transform=do_nothing,
                    get_kw=None, do_refresh=True):
    update = dict(body_transform(update))
    response = endpoint.put(
        json=({endpoint.form_key: update}
              if endpoint.form_key is not None else update)
    )

    if response.status_code in range(200, 300):
        if do_refresh:
            return endpoint.refresh(**(get_kw or {}))
        return endpoint

    log.error(
        f'There was an error after updating endpoint {endpoint.url}:\n'
        '\n'
        f'Update dict: \n{pprint.pformat(update)}\n'
        '\n'
        'Response:\n'
        f'{response.content[:1000]}'
    )
    return endpoint


# ----------------------------------------------------------------------
# Caching logic
#
#
# It is assumed that each type of Endpoint has lists of other
# Endpoints associated with them. For a given parent Endpoint and a
# child resource name, the cache stores a list of child objects.
#
# ----------------------------------------------------------------------

_cache = {}
def memoize_key(parent_endpoint: Endpoint, resource_name: str):
    return (parent_endpoint.url, resource_name)

@curry
def memoize_resources(parent_endpoint: Endpoint, resource_name: str,
                      resources: List[Endpoint]):
    global _cache
    _cache[memoize_key(parent_endpoint, resource_name)] = resources
    return resources

def cache_get(parent_endpoint: Endpoint, resource_name: str, default=None):
    return _cache.get(memoize_key(parent_endpoint, resource_name), default)
    
def cache_remove(parent_endpoint: Endpoint, resource_name: str, default=None):
    return cache_remove_by_key(
        memoize_key(parent_endpoint, resource_name), default,
    )
    
def cache_remove_by_key(key, default=None):
    return _cache.pop(key, default)
    
def cache_has_key(parent_endpoint: Endpoint, resource_name: str):
    return memoize_key(parent_endpoint, resource_name) in _cache

def total_cache_reset():
    log.debug('TOTAL CACHE RESET')
    global _cache
    _cache = {}

def reset_cache_by_endpoint(parent_endpoint: Endpoint):
    for (url, resource_name) in tuple(_cache):
        if parent_endpoint.url == url:
            log.debug(f'Reset {resource_name} cache for {url}')
            cache_remove(parent_endpoint, resource_name)

def reset_cache_by_resource_name(resource_name: str):
    for (url, name) in tuple(_cache):
        if name == resource_name:
            log.debug(f'Reset {resource_name} cache for {url}')
            cache_remove_by_key((url, name))

def reset_cache_for_endpoint_by_resource_name(parent_endpoint: Endpoint,
                                              resource_name: str):
    cache_remove(parent_endpoint, resource_name)

def get_id_resource(resource_name: str, *,
                    form_key: str = None,
                    id_key: str = 'id', meta_f=empty_dict,
                    unpack_f=do_nothing, single_unpack_f=do_nothing,
                    help=None, **iter_kw):
    def getter(parent_endpoint: IdResourceEndpoint, id: (int, str), **get_kw):
        return pipe(
            parent_endpoint(resource_name, id).get(**get_kw),
            IdResourceEndpoint.from_single_response(
                parent_endpoint,
                form_key=form_key, id_key=id_key, unpack_f=unpack_f,
                meta_f=meta_f, single_unpack_f=single_unpack_f,
            )
        )
    getter.__doc__ = help or ''

    return getter

def get_id_resources(resource_name: str, *, form_key: str = None,
                     id_key: str = 'id', meta_f=empty_dict,
                     unpack_f=do_nothing, single_unpack_f=do_nothing,
                     help=None, memo=False, **iter_kw):
    def getter(parent_endpoint: IdResourceEndpoint, *, do_memo=True):
        if (memo and do_memo) and cache_has_key(parent_endpoint,
                                                resource_name):
            return cache_get(parent_endpoint, resource_name)

        return pipe(
            parent_endpoint(resource_name).iter(
                'get', **merge(
                    {'iter_f': compose(
                        IdResourceEndpoint.from_multiple_response(
                            form_key=form_key, id_key=id_key,
                            meta_f=meta_f, unpack_f=unpack_f,
                            single_unpack_f=single_unpack_f,
                        ),
                    )},
                    iter_kw
                )
            ),
            tuple,
            memoize_resources(parent_endpoint, resource_name),
        )
    getter.__doc__ = help or ''

    getter.reset_cache = partial(
        reset_cache_for_endpoint_by_resource_name,
        resource_name=resource_name,
    )
    getter.reset_cache.__doc__ = f'''
    Reset the "{resource_name}" cache for a given Endpoint.
    '''

    return getter

def new_id_resource(resource_name: str, *,
                    form_key: str = None, id_key: str = 'id',
                    get_kw=None, help: str = None, memo=False,
                    meta_f=empty_dict, unpack_f=do_nothing,
                    single_unpack_f=do_nothing,
                    post_unpack_f=do_nothing,
                    body_transform=do_nothing):
    def creator(parent_endpoint: Endpoint, body: dict):
        body = dict(body_transform(body))
        body = {form_key: body} if form_key is not None else body
        log.debug(f'Request body: {body}')
        response = parent_endpoint(resource_name).post(json=body)

        if response.status_code in range(200, 300):
            post_data = post_unpack_f(response.json())
            new_response = parent_endpoint(
                resource_name, post_data[id_key]
            ).get(**(get_kw or {}))
            if new_response.status_code in range(200, 300):
                cache_remove(parent_endpoint, resource_name)
                return IdResourceEndpoint(
                    parent_endpoint(resource_name),
                    new_response.json(),
                    form_key=form_key, id_key=id_key,
                    meta_f=meta_f,
                    unpack_f=unpack_f, single_unpack_f=single_unpack_f,
                )

            log.error(
                f'There was an error after creating {resource_name} object:\n'
                f'  body: {body}\n'
                f'  form_key: {form_key}\n'
                f'  id_key: {id_key}\n'
                'Response:\n'
                f'{new_response.content[:1000]}'
            )

        log.error(
            f'There was an error creating {resource_name} object:\n'
            f'  body: {body}\n'
            f'  form_key: {form_key}\n'
            f'  id_key: {id_key}\n'
            'Response:\n'
            f'{response.content[:1000]}'
        )
    creator.__doc__ = help or ''
    return creator




import os
import io
import traceback
import importlib
import logging
from ipaddress import ip_address, ip_interface, ip_network
from typing import (
    Iterable, Hashable, Union, Sequence, Any, Callable, Tuple,
)
import csv
from pathlib import Path
import builtins
import collections
import functools
import re
import random
import inspect
import math
import textwrap
import collections.abc
from collections import OrderedDict
from datetime import datetime
import json
import tempfile                 # noqa: for doctest
import base64
import urllib
import string

from multipledispatch import dispatch
from pyrsistent import pmap, pvector, PVector
import netifaces
import jmespath
import dateutil.parser
import requests

try:
    from cytoolz.curried import (
        curry, pipe, compose, merge, concatv,
        map, mapcat, assoc, dissoc, valmap, first, second, last,
        complement, get as _get, concat, filter, do, groupby,
        partial, juxt,
    )
except ImportError:
    from toolz.curried import (
        curry, pipe, compose, merge, concatv,
        map, mapcat, assoc, dissoc, valmap, first, second, last,
        complement, get as _get, concat, filter, do, groupby,
        partial, juxt,
    )

log = logging.getLogger('common')
log.addHandler(logging.NullHandler())

ip_re = re.compile(
    r'(?:(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])\.){3}'
    r'(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])(?![\d\.]+)'
)
ip_only_re = re.compile(f'^{ip_re.pattern}$')

@curry
def log_lines(log_function, lines):
    return pipe(
        lines,
        mapcat(lambda line: line.splitlines()),
        filter(None),
        map(log_function),
    )

def lines(path: Union[str, Path]):
    return Path(path).expanduser().read_text().splitlines()

def do_nothing(value):
    '''Yes, we have no banana pudding.

    Examples:

    >>> do_nothing(1)
    1
    >>> do_nothing("Banana Pudding")
    'Banana Pudding'
    '''
    return value

def b64decode(content: Union[bytes, str]):
    return base64.b64decode(
        to_bytes(content) + b'=' * (len(content) % 4)
    )

def b64encode(content: Union[bytes, str]):
    return pipe(
        content,
        to_bytes,
        base64.b64encode,
    )
    
# ----------------------------------------------------------------------
#
# Random content creation functions
#
# ----------------------------------------------------------------------

def random_str(n=8, *, rng=None):
    rng = rng or random
    return ''.join(rng.sample(string.ascii_lowercase, n))

def random_sentence(w=10, *, rng=None):
    rng = rng or random
    return pipe(
        [random_str(rng.randrange(4, 10), rng=rng) for i in range(w)],
        lambda s: [s[0].capitalize()] + s[1:],
        ' '.join,
        lambda s: s + '.'
    )

def random_user(n=8, *, rng=None):
    rng = rng or random
    return ''.join(rng.sample(string.ascii_lowercase, n))

def random_pw(n=16, *, rng=None):
    rng = rng or random
    return ''.join(rng.sample(string.printable[:64], n))

@curry
def random_sample(N, seq, *, rng=None):
    rng = rng or random
    return rng.sample(tuple(seq), N)


# ----------------------------------------------------------------------
#
# HTTP functions
#
# ----------------------------------------------------------------------

def url(*parts):
    base, *path = pipe(parts, map(str))
    return urllib.parse.urljoin(
        base, '/'.join(path),
        # base, Path(*path).as_posix(),
    )

def session_with_cookies(cookies: Tuple[dict]):
    session = requests.Session()
    for cookie_dict in cookies:
        session.cookies.set_cookie(
            requests.cookies.create_cookie(**cookie_dict)
        )
    return session

def valid_response(response: requests.Response):
    return response.status_code in range(200, 300)

def valid_content(response: requests.Response):
    return response.content if valid_response(response) else Null

# ----------------------------------------------------------------------
#
# pyrsistent object functions
#
# ----------------------------------------------------------------------

def to_pyrsistent(obj):
    '''Convert object to immutable pyrsistent objects

    Examples:

    >>> to_pyrsistent({'a': 1})
    pmap({'a': 1})
    
    >>> to_pyrsistent({'a': [1, 2, 3]})['a']
    pvector([1, 2, 3])
    
    >>> to_pyrsistent({'a': [1, 2, 3]})['a'][0] = 2
    Traceback (most recent call last):
      ...
    TypeError: 'pvectorc.PVector' object does not support item assignment
    '''
    # return pyrsistent.freeze(obj)
    if is_dict(obj):
        return pipe(
            obj.items(),
            vmap(lambda k, v: (k, to_pyrsistent(v))),
            pmap,
        )
    if is_seq(obj):
        return pipe(obj, map(to_pyrsistent), pvector)
    return obj

def no_pyrsistent(obj):
    '''Convert all pyrsistent objects to Python types

    pmap -> dict
    pvector -> tuple

    Examples:

    >>> pipe(pmap({'a': pvector([1, 2, 3])}), no_pyrsistent)
    {'a': (1, 2, 3)}
    '''
    # return pyrsistent.thaw(obj)
    if is_dict(obj):
        return pipe(
            obj.items(),
            vmap(lambda k, v: (k, no_pyrsistent(v))),
            dict,
        )
    if is_seq(obj):
        return pipe(obj, map(no_pyrsistent), tuple)
    return obj

def freeze(func):
    '''Ensure output of func is immutable

    Uses to_pyrsistent on the output of func

    Examples:
    
    >>> @freeze
    ... def f():
    ...     return [1, 2, 3]
    >>> f()
    pvector([1, 2, 3])
    '''
    @functools.wraps(func)
    def return_frozen(*a, **kw):
        return pipe(func(*a, **kw), to_pyrsistent)
    return return_frozen

frozen_curry = compose(curry, freeze)


# ----------------------------------------------------------------------
#
# Error handling functions
#
# ----------------------------------------------------------------------

def mini_tb(levels=3):
    '''Traceback message suitable for logging

    '''
    frame = inspect.currentframe().f_back
    parents = [frame.f_back]
    for i in range(levels - 1):
        if parents[-1].f_back:
            parents.append(parents[-1].f_back)
        else:
            break
    return '\n' + pipe(
        parents,
        map(inspect.getframeinfo),
        vmap(lambda filen, lnum, fname, lns, i: (
            f'{Path(filen).name}', lnum, fname, lns[i],
        )),
        vmap(lambda path, lnum, fname, line: (
            f'- {fname} | {path}:{lnum} | {line.rstrip()}'
        )),
        '\n'.join,
    )


# ----------------------------------------------------------------------
#
# JSON handling functions
#
# ----------------------------------------------------------------------

def maybe_json(response: requests.Response):
    try:
        return response.json()
    except ValueError:
        return Null

@curry
@functools.wraps(json.dumps)
def json_dumps(*a, **kw):
    return json.dumps(*a, **kw)

@curry
@functools.wraps(json.loads)
def json_loads(*a, **kw):
    return json.loads(*a, **kw)

@curry
def jmes(search, d):
    '''Curried jmespath.search

    Examples:

    >>> pipe({'a': {'b': [10, 9, 8]}}, jmes('a.b[2]'))
    8
    '''
    if is_null(d):
        log.error(
            f'null dict passed to jmes {mini_tb(5)}'
        )
        return Null
    return jmespath.search(search, d)


# ----------------------------------------------------------------------
#
# Builtin function/object supplements
#
# ----------------------------------------------------------------------

@curry
def cprint(value_func=do_nothing, **print_kw):
    def do_print(value, **kw):
        return print(value_func(value), **merge(print_kw, kw))
    return do_print

@curry
def max(iterable, **kw):
    '''Curried max

    Examples:

    >>> pipe([5, 2, 1, 10, -1], max())
    10
    >>> pipe([5, 2, 1, 10, -1], max(key=lambda v: 1 / v))
    1
    '''
    return builtins.max(iterable, **kw)

@curry
def min(iterable, **kw):
    '''Curried min

    Examples:

    >>> pipe([5, 2, 1, 10, 4], min())
    1
    >>> pipe([5, 2, 1, 10, 4], min(key=lambda v: 1 / v))
    10
    '''
    return builtins.min(iterable, **kw)

@curry
@functools.wraps(builtins.sorted)
def sorted(iterable, **kw):
    '''Curried sorted

    Examples:

    >>> pipe([5, 2, 6], sorted)
    [2, 5, 6]
    '''
    return builtins.sorted(iterable, **kw)

@curry
def sort_by(func, iterable, **kw):
    '''Sort iterable by key=func

    Examples:

    >>> pipe([{'a': 1}, {'a': 8}, {'a': 2}], sort_by(getitem('a')))
    [{'a': 1}, {'a': 2}, {'a': 8}]
    '''
    return builtins.sorted(iterable, key=func, **kw)

def cat_to_set(iterable):
    '''Concatenate all iterables in iterable to single set

    Examples:

    >>> the_set = pipe([(1, 2), (3, 2), (2, 8)], cat_to_set)
    >>> the_set == {1, 2, 3, 8}
    True
    '''
    result = set()
    for iterable_value in iterable:
        result.update(iterable_value)
    return result

@curry
def map_to_set(func: Callable[[Any], Hashable], iterable):
    '''Map func over iterable, reduce result to set.

    Return value of func needs to be hashable.

    Examples:

    >>> the_set = pipe([(1, 2), (3, 2), (2, 8)], map_to_set(lambda v: v[0]))
    >>> the_set == {1, 2, 3}
    True
    '''
    result = set()
    for value in iterable:
        result.add(func(value))
    return result

_getattr = builtins.getattr
@curry
def deref(attr, obj):
    '''Curried derefrencing function for accessing attributes of an
    object.

    Examples:

    >>> class X:
    ...     def __init__(self, x):
    ...         self.x = x
    ...
    >>> pipe([X(1), X(5)], map(deref('x')), tuple)
    (1, 5)
    '''
    return _getattr(obj, attr)

def call(method_name, *a, **kw):
    '''"Curried" method caller

    Examples:

    >>> class X:
    ...     def __init__(self, x):
    ...         self.x = x
    ...     def square(self):
    ...         return self.x ** 2
    ...     def mult(self, v):
    ...         return self.x * v
    ...
    >>> pipe([X(1), X(5)], map(call('square')), tuple)
    (1, 25)
    >>> pipe([X(1), X(5)], map(call('mult', 3)), tuple)
    (3, 15)
    '''
    def caller(obj):
        return _getattr(obj, method_name)(*a, **kw)
    return caller

# ----------------------------------------------------------------------
#
# Filesystem functions
#
# ----------------------------------------------------------------------

def walk(path):
    '''Return os.walk(path) as sequence of Path objects

    >>> with tempfile.TemporaryDirectory() as temp:
    ...     root = Path(temp)
    ...     Path(root, 'a', 'b').mkdir(parents=True)
    ...     _ = Path(root, 'a', 'a.txt').write_text('')
    ...     _ = Path(root, 'a', 'b', 'b.txt').write_text('')
    ...     paths = tuple(walk(root))
    >>> paths == (Path(root, 'a', 'a.txt'), Path(root, 'a', 'b', 'b.txt'))
    True

    '''
    return pipe(
        os.walk(path),
        vmapcat(lambda root, dirs, files: [Path(root, f) for f in files]),
    )

@curry
def walkmap(func, root):
    '''Map function over all paths in os.walk(root)

    '''
    return pipe(
        walk(root),
        map(func),
    )

def check_parents_for_file(name, start_dir=Path('.')):
    start_path = Path(start_dir).expanduser()
    directories = concatv([start_path], start_path.parents)
    for base in directories:
        path = Path(base, name)
        if path.exists():
            return path
    return Null

def to_paths(*paths):
    return pipe(paths, map(Path), tuple)

@curry
def newer(path: Union[str, Path], test: Union[str, Path]):
    '''Is the path newer than the test path?

    '''
    return ctime(path) > ctime(test)

@curry
def older(path: Union[str, Path], test: Union[str, Path]):
    '''Is the path older than the test path?

    '''
    return ctime(path) < ctime(test)


# ----------------------------------------------------------------------
#
# CSV functions
#
# ----------------------------------------------------------------------

@curry
def csv_rows_from_path(path: Union[str, Path], *, header=True,
                       columns=None, **kw):
    '''Load CSV rows from file path

    '''
    return csv_rows_from_fp(
        Path(path).expanduser().open(), header=header,
        columns=columns, **kw
    )
csv_rows = csv_rows_from_path

@curry
def csv_rows_from_content(content: Union[str, bytes], *,
                          header=True, columns=None, **kw):
    r'''Load CSV rows from content (e.g. str or bytes)

    Args:
      content (str): string content

    Examples:
    
    >>> pipe(csv_rows_from_content('c1,c2,c3\n1,2,3'), list)
    [OrderedDict([('c1', '1'), ('c2', '2'), ('c3', '3')])]

    If header is False, then rows will be returned as lists.
    
    >>> pipe(csv_rows_from_content('1,2,3\n4,5,6', header=False), list)
    [['1', '2', '3'], ['4', '5', '6']]

    >>> pipe(csv_rows_from_content(
    ...   '1,2,3', header=False, columns=['c1', 'c2', 'c3']
    ... ), list)
    [OrderedDict([('c1', '1'), ('c2', '2'), ('c3', '3')])]

    If header is False and header row exists, the header row will be
    interpreted as a regular row.
    
    >>> pipe(csv_rows_from_content('c1,c2,c3\n1,2,3', header=False), list)
    [['c1', 'c2', 'c3'], ['1', '2', '3']]

    '''
    return csv_rows_from_fp(
        io.StringIO(content), header=header, columns=columns, **kw
    )

@curry
def csv_rows_from_fp(rfp, *, header=True, columns=None, **reader_kw):
    '''Load CSV rows from file-like object

    '''
    if header:
        column_row = next(csv.reader(rfp))
        columns = columns or column_row
        reader = csv.DictReader(rfp, columns, **reader_kw)
    elif is_seq(columns):
        reader = csv.DictReader(rfp, columns, **reader_kw)
    else:
        reader = csv.reader(rfp, **reader_kw)
    for row in pipe(reader, filter(None)):
        yield row
    
@curry
def csv_rows_to_fp(wfp, rows: Iterable[Union[dict, Sequence[str]]], *,
                   header: bool = True,
                   columns: Union[dict, Iterable[str]] = None,
                   **writer_kw):
    r'''Save CSV rows to file-like object

    Args:

      wfp (file-like): File-like object into which to write the CSV
        content

      rows (Iterable[Union[dict, Iterable]]): Row data to write to
        CSV.

        Iterable[dict], columns is None: If given as iterable of
        dictionaries and columns is None, columns will come from keys
        of row dictionaries. This means that __the row data will need
        to be exhausted__ to build column list. The final column
        sequence will be sorted.

        Iterable[dict], columns is dict: If given as iterable of
        dictionaries and columns is a dictionary, then it is assumed
        to be a mapping from row keys to the final columns. If final
        column ordering is important, then use a
        collections.OrderedDict to encode the columns.

        Iterable[dict], columns is Iterable[str]: If given as iterable
        of dictionaries and columns is an iterable, then it will be
        used as the final list of columns. It is __assumed that the
        iterable of columns contains all necessary columns__. Only the
        given columns will be provided in the final CSV data.

        Iterable[Sequence[str]], columns is None: If given as iterable
        of sequences and columns is None, then there will be no header
        in the final CSV.

        Iterable[Sequence[str]], columns is Iterable[str]: If given as
        iterable of sequences and columns is an iterable, then there
        will be a header in the final CSV if header is True.

      header (bool): Should there be a header in the final CSV?

      columns (Union[dict, Iterable[str]]): Columns to be used in
        final CSV

      **writer_kw: Keyword arguments to be passed to csv.writer (or
        csv.DictWriter)

    Examples:

    >>> import io
    >>> from pathlib import Path
    >>> wfp = io.StringIO()
    >>> pipe(
    ...     [{'a': 1, 'b': 2}, {'a': 3, 'b': 4}],
    ...     csv_rows_to_fp(wfp),
    ... )
    >>> wfp.getvalue() == 'a,b\r\n1,2\r\n3,4\r\n'
    True
    >>> wfp = io.StringIO()
    >>> pipe(
    ...     [{'a': 1, 'b': 2}, {'a': 3, 'b': 4}],
    ...     csv_rows_to_fp(wfp, columns={'b': 'B', 'a': 'A'}),
    ... )
    >>> assert wfp.getvalue() in {'A,B\r\n1,2\r\n3,4\r\n', 'B,A\r\n2,1\r\n4,3\r\n'}
    >>> wfp = io.StringIO()
    >>> pipe(
    ...     [(1, 2), (3, 4)],
    ...     csv_rows_to_fp(wfp, columns=['a', 'b']),
    ... )
    >>> assert wfp.getvalue() == 'a,b\r\n1,2\r\n3,4\r\n'
    >>> wfp = io.StringIO()
    >>> pipe(
    ...     [(1, 2), (3, 4)],
    ...     csv_rows_to_fp(wfp),
    ... )
    >>> assert wfp.getvalue() == '1,2\r\n3,4\r\n'

    '''
    
    row_iter = iter(rows)
    first_row = next(row_iter)
    # If rows are passed as iterable of sequences, each row must be an
    # in-memory sequence like a list, tuple, or pvector (i.e. not an
    # iter or generator), otherwise, this will have the
    # __side-effect__ of exhausting the first row.
    rows_are_dicts = is_dict(first_row)
    columns_is_dict = is_dict(columns)

    rows = concatv([first_row], row_iter)
    if rows_are_dicts:
        if columns_is_dict:
            items = tuple(columns.items())
            rows = pipe(
                rows,
                map(lambda r: OrderedDict([
                    (to_c, r[from_c]) for from_c, to_c in items
                ])),
            )
            columns = list(columns.values())
        elif columns is None:
            rows = tuple(rows)
            columns = pipe(
                rows,
                map(call('keys')),
                cat_to_set,
                sorted,
            )
        else:                   # assuming columns is Iterable
            columns = tuple(columns)
            rows = pipe(
                rows,
                map(lambda r: {
                    c: r[c] for c in columns
                }),
            )
        writer = csv.DictWriter(wfp, columns, **writer_kw)
        if header:
            writer.writeheader()
    else:                       # assuming rows are Iterable
        if columns is not None:  # assuming columns is Iterable
            columns = tuple(columns)
            rows = pipe(
                rows,
                map(lambda r: {
                    c: r[i] for i, c in enumerate(columns)
                }),
            )
            writer = csv.DictWriter(wfp, columns)
            if header:
                writer.writeheader()
        else:
            writer = csv.writer(wfp, **writer_kw)
            
    writer.writerows(rows)

@curry
def csv_rows_to_path(path: Union[str, Path],
                     rows: Iterable[Union[dict, Sequence[str]]], *,
                     header: bool = True,
                     columns: Union[dict, Iterable[str]] = None,
                     **writer_kw):
    '''Save CSV rows to file system path

    '''
    with Path(path).expanduser().open('w') as wfp:
        return csv_rows_to_fp(
            wfp, rows, header=header, columns=columns, **writer_kw
        )

@curry
def csv_rows_to_content(rows: Iterable[Union[dict, Sequence[str]]], *,
                        header: bool = True,
                        columns: Union[dict, Iterable[str]] = None,
                        **writer_kw):
    '''Save CSV rows to a string

    '''
    buf = io.StringIO()
    csv_rows_to_fp(
        buf, rows, header=header, columns=columns, **writer_kw
    )
    return buf.getvalue()

# ----------------------------------------------------------------------
#
# Supplemental versions of toolz functions, especially variadic
# versions.  Also some additional toolz-like functions.
#
# ----------------------------------------------------------------------

@curry
def vcall(func, value):
    '''Variadic call

    Example:

    >>> vcall(lambda a, b: a + b)([1, 2])
    3
    '''
    return func(*value)

@curry
def vmap(func, seq):
    '''Variadic map

    Example:

    >>> pipe([(2, 1), (2, 2), (2, 3)], vmap(lambda a, b: a ** b), tuple)
    (2, 4, 8)
    '''
    return (func(*v) for v in seq)
    
starmap = vmap

@curry
def vfilter(func, seq):
    '''Variadic filter

    Example:

    >>> pipe([(1, 2), (4, 3), (5, 6)], vfilter(lambda a, b: a < b), tuple)
    ((1, 2), (5, 6))
    '''
    return filter(vcall(func), seq)

@curry
def vmapcat(func, seq):
    '''Variadic mapcat

    Example:

    >>> pipe([(1, 2), (4, 3), (5, 6)], vmapcat(lambda a, b: [a] * b), tuple)
    (1, 1, 4, 4, 4, 5, 5, 5, 5, 5, 5)
    '''
    return pipe(concat(func(*v) for v in seq), tuple)

@curry
def vgroupby(key_func, seq):
    return groupby(vcall(key_func), seq)

@curry
def vvalmap(val_func, d, factory=dict):
    return valmap(vcall(val_func), d, factory=factory)

@curry
def select(keys, iterable):
    '''Select a set of keys out of each indexable in iterable. Assumes
    that these keys exist in each indexable (i.e. will throw
    IndexError or KeyError if they don't)

    Example:

    >>> pipe([{'a': 1, 'b': 2}, {'a': 3, 'b': 4}], select(['a', 'b']), tuple)
    ((1, 2), (3, 4))

    '''
    for indexable in iterable:
        yield tuple(indexable[k] for k in keys)

@curry
def find(find_func, iterable):
    '''Finds first value in iterable that when passed to find_func returns
    truthy

    Example:

    >>> pipe([0, 0, 1, -1, -2], find(lambda v: v < 0))
    -1
    '''
    for value in iterable:
        if find_func(value):
            return value
    return Null

@curry
def vfind(find_func, iterable):
    '''Variadic find: finds first value in iterable that when passed to
    find_func returns truthy

    Example:

    >>> pipe([(0, 0), (0, 1)], vfind(lambda a, b: a < b))
    (0, 1)

    '''
    return find(vcall(find_func), iterable)

@curry
def index(find_func, iterable):
    '''Finds index of the first value in iterable that when passed to
    find_func returns truthy

    Example:

    >>> pipe([0, 0, 1, -1, -2], index(lambda v: v < 0))
    3

    '''
    for i, value in enumerate(iterable):
        if find_func(value):
            return i
    return Null

@curry
def vindex(find_func, iterable):
    '''Variadic vindex: finds index of first value in iterable that when
    passed to find_func returns truthy

    Example:

    >>> pipe([(0, 0), (0, 1)], vindex(lambda a, b: a < b))
    1

    '''
    return index(vcall(find_func), iterable)

@curry
def seti(index, func, iterable):
    '''Return copy of iterable with value at index modified by func

    Examples:

    >>> pipe([10, 5, 2], seti(2, lambda v: v**2), list)
    [10, 5, 4]
    '''
    for i, v in enumerate(iterable):
        if i == index:
            yield func(v)
        else:
            yield v
seti_t = compose(tuple, seti)

@curry
def vseti(index, func, iterable):
    '''Variadict seti: return iterable of seq with value at index modified
    by func

    Examples:

    >>> pipe(
    ...   [(10, 1), (5, 2), (2, 3)],
    ...   vseti(2, lambda v, e: v**e),
    ...   list
    ... )
    [(10, 1), (5, 2), 8]
    '''
    return seti(index, vcall(func), iterable)
vseti_t = compose(tuple, vseti)

@curry
def callif(if_func, func, value):
    '''Return func(value) only if if_func(value) returns truthy, otherwise
    Null

    Examples:

    >>> str(callif(lambda a: a > 0, lambda a: a * a)(-1))
    'Null'
    >>> callif(lambda a: a > 0, lambda a: a * a)(2)
    4
    '''
    if if_func(value):
        return func(value)
    return Null

@curry
def vcallif(if_func, func, value):
    '''Variadic callif: return func(value) only if if_func(value) returns
    truthy, otherwise Null. Both if_func and func are called
    variadically.

    Examples:

    >>> str(callif(lambda a: a > 0, lambda a: a * a)(-1))
    'Null'
    >>> callif(lambda a: a > 0, lambda a: a * a)(2)
    4

    '''
    return callif(vcall(if_func), vcall(func), value)

@curry
def vdo(func, value):
    '''Variadic do

    '''
    return do(vcall(func), value)

@curry
def mapdo(do_func, iterable):
    '''Map a do function over values in iterable. It will generate all
    values in iterable immediately, run do_func over those values, and
    return the values as a tuple (i.e. there are **side effects**).

    Examples:

    >>> l = [0, 1, 2]
    >>> pipe(l, mapdo(print))
    0
    1
    2
    (0, 1, 2)
    >>> l is pipe(l, mapdo(do_nothing))
    False

    '''
    values = tuple(iterable)
    for v in values:
        do_func(v)
    return values

@curry
def vmapdo(do_func, iterable):
    return mapdo(vcall(do_func), iterable)

@curry
def mapif(func, seq):
    # return [func(*v) for v in seq]
    if func:
        return (func(v) for v in seq)
    return seq
        
@curry
def vmapif(func, seq):
    # return [func(*v) for v in seq]
    if func:
        return (func(*v) for v in seq)
    return seq

@curry
def groupdicts(regex, iterable, **kw):
    regex = regex if isinstance(regex, re.Pattern) else re.compile(regex, **kw)
    return pipe(
        iterable,
        map(regex.search),
        filter(None),
        map(lambda m: m.groupdict()),
    )

@curry
def groupdicts_from_regexes(regexes: Sequence[re.Pattern],
                            iterable: Iterable[str], **kw):
    return pipe(
        iterable,
        juxt(*[compose(list, groupdicts(r)) for r in regexes]),
        concat,
    )

@curry
def grep(raw_regex, iterable, **kw):
    regex = re.compile(raw_regex, **kw)
    return filter(lambda s: regex.search(s), iterable)

@curry
def grepv(raw_regex, iterable, **kw):
    regex = re.compile(raw_regex, **kw)
    return filter(lambda s: not regex.search(s), iterable)

@curry
def grepitems(raw_regex, iterable, **kw):
    regex = re.compile(raw_regex, **kw)
    return pipe(
        iterable,
        filter(lambda items: any(regex.search(s) for s in items)),
        tuple,
    )

@curry
def grepvitems(raw_regex, iterable, **kw):
    regex = re.compile(raw_regex, **kw)
    return pipe(
        iterable,
        filter(lambda items: not any(regex.search(s) for s in items)),
        tuple,
    )

def shuffled(seq):
    tup = tuple(seq)
    return random.sample(tup, len(tup))

def first_true(iterable, *, default=None):
    '''Return the first truthy thing in iterable. If none are true, return
    default=Null.

    '''
    for v in iterable:
        if v:
            return v
    return Null if default is None else default

@curry
def getitem(i, indexable, default=None):
    default = default or Null
    if is_dict(indexable):
        return indexable.get(i, default)
    return _get(i, indexable, default)
get = getitem

@curry
def get_many(keys, indexable, default=None):
    '''Get multiple keys/indexes from an indexable object

    Example:

    >>> pipe([2, 6, 1, 5, 8, -3], getmany([0, 5]), tuple)
    (2, -3)
    '''
    for k in keys:
        yield get(k, indexable, default=default)
getmany = get_many

@curry 
def get_many_t(keys, indexable, default=None):
    return pipe(
        get_many(keys, indexable, default=default),
        tuple,
    )
getmany_t = get_many_t

# ----------------------------------------------------------------------
#
# Monad(ish) functions (e.g. approximation to Maybe monad)
#
# ----------------------------------------------------------------------

class _null:
    '''Null type for creating pseudo-monads.

    Similar to Nothing in Haskell

    Do **not** use as an iterable (i.e. in for loops or over maps), as
    this leads to **infinite loops**.

    '''
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(_null, cls).__new__(cls)
        return cls._instance

    def __repr__(self):
        return 'Null'

    def __bool__(self):
        return False

    def __len__(self):
        return 0

    def __contains__(self, *a):
        return False

    def __next__(self):
        return Null

    def __getattr__(self, key):
        if key == '__wrapped__':
            return lambda *a, **kw: None
        return Null

    def __getitem__(self, key):
        return Null

    def __lt__(self, other):
        return False

    def __gt__(self, other):
        return False

    def __eq__(self, other):
        return False

    def __le__(self, other):
        return False

    def __ge__(self, other):
        return False

    def __call__(self, *a, **kw):
        return Null

    def __add__(self, *a):
        return Null

    def __radd__(self, *a):
        return Null

    def __sub__(self, *a):
        return Null

    def __rsub__(self, *a):
        return Null

    def __mul__(self, *a):
        return Null

    def __rmul__(self, *a):
        return Null

    def __div__(self, *a):
        return Null

    def __rdiv__(self, *a):
        return Null

Null = _null()

def is_null(v):
    return v is None or v is Null
not_null = complement(is_null)

def maybe(value, default=None):
    '''If value is "null" (i.e. is either None or the Null object), return
    Null, otherwise return value. Like Scala's ?? operator.

    Examples:

    >>> maybe({'a': [0, {'b': 1}]}.get('a'))[1]['b']
    1
    >>> maybe({}.get('a'))[1]['b']
    Null

    '''
    if is_null(value):
        if default is not None:
            return default
        return Null
    return value

def maybe_pipe(value, *functions, default=None):
    '''Sort-of Maybe monad. Pipe value through series of functions unless
    and until one of them returns a "null" (i.e. None or Null) value
    or throws an Exception, where it will return Null or a non-null
    default value)

    '''
    if is_null(value):
        return Null
    for f in functions:
        try:
            value = f(value)
        except Exception:
            log.error(f'Error in maybe_pipe: \n{traceback.format_exc()}')
            return Null if default is None else default
        if is_null(value):
            return Null if default is None else default
    return value

def maybe_int(value, default=None):
    '''Convert to int or return Null (or non-null default)

    '''
    if is_int(value):
        return int(value)
    return default or Null

def is_int(value):
    try:
        int(value)
    except ValueError:
        return False
    except TypeError:
        return False
    return True

def maybe_float(value, default=None):
    '''Convert to float or return Null (or non-null default)

    '''
    if is_float(value):
        return float(value)
    return default or Null

def is_float(value):
    try:
        float(value)
    except ValueError:
        return False
    except TypeError:
        return False
    return True

@curry
def maybe_max(iterable, **kw):
    '''Return max of iterable or (if empty) return Null

    '''
    try:
        return max(iterable, **kw)
    except ValueError:
        return Null

@curry
def maybe_min(iterable, **kw):
    '''Return min of iterable or (if empty) return Null

    '''
    try:
        return min(iterable, **kw)
    except ValueError:
        return Null

@curry
def short_circuit(function, value):
    '''If function(value) is falsy, return Null. Useful for
    inserting into maybe_pipe to short-circuit.

    Different from maybe in that maybe is specifically for "null"
    values, not falsy things.

    '''
    if not function(value):
        return Null
    return value

def sc_juxt(*funcs):
    '''Short-circuiting juxt

    '''
    def caller(*a, **kw):
        sc = False
        for f in funcs:
            if sc:
                yield Null
            output = f(*a, **kw)
            if not output:
                sc = True
                yield Null
            else:
                yield output
    caller.__doc__ = help_text(f'''
    Juxtaposition of {pipe(funcs, map(deref('__name__')), ', '.join)}.
    Will short-circuit on the first falsy return value and return Nulls
    thereafter.
    ''')
    return caller

def maybe_first(iterable, *, default=None):
    '''Return first element of iterable. If empty return default=Null.

    '''
    try:
        return first(iterable)
    except StopIteration:
        pass
    return Null if default is None else default

def maybe_second(iterable, *, default=None):
    '''Return second element of iterable. If empty return default=Null.

    '''
    try:
        return second(iterable)
    except StopIteration:
        pass
    return Null if default is None else default

def maybe_last(iterable, *, default=None):
    '''Return last element of iterable If empty return default=Null.

    '''
    try:
        return last(iterable)
    except StopIteration:
        pass
    return Null if default is None else default


# ----------------------------------------------------------------------
#
# Dictionary functions
#
# ----------------------------------------------------------------------

def cmerge(*dicts):
    '''Curried dictionary merge

    Examples:

    >>> merged = pipe({'a': 1}, cmerge({'b': 2}, {'c': 3}))
    >>> merged == {'a': 1, 'b': 2, 'c': 3}
    True

    '''
    def do_merge(*more_dicts):
        return merge(*(dicts + more_dicts))
    return do_merge

@curry
def create_key(key: Hashable, value_function, d):
    '''Create key in a given dictionary if it doesn't already exist

    Args:
      key (Hashable): key to be added (if it doesn't already exist)
    
      value_function (Callable[[dict], Any]): function that takes the
        dictionary and returns a value for the new key

      d (dict): dictionary to transform (with no side effects)

    Returns: (dict) new state of dictionary

    Examples:

    >>> new = pipe({'b': 2}, create_key('a', lambda d: d['b'] + 10))
    >>> new == {'a': 12, 'b': 2}
    True
    >>> new = pipe({'a': 1, 'b': 2}, create_key('a', lambda d: d['b'] + 10))
    >>> new == {'a': 1, 'b': 2}
    True

    '''
    if key not in d:
        return assoc(d, key, value_function(d))
    return d

@curry
def update_key(key: Hashable, value_function, d):
    '''Update key's value for a given dictionary. Will add key if it
    doesn't exist. Basically just a curried version of assoc.

    Args:
      key (Hashable): key to be updated
    
      value_function (Callable[[dict], Any]): function that takes the
        dictionary and returns a value for the key

      d (dict): dictionary to transform (with no side effects)

    Returns: (dict) new state of dictionary

    Examples:

    >>> new = pipe({'b': 2}, update_key('a', lambda d: d['b'] + 10))
    >>> new == {'a': 12, 'b': 2}
    True
    >>> new = pipe({'a': 1, 'b': 2}, update_key('a', lambda d: d['b'] + 10))
    >>> new == {'a': 12, 'b': 2}
    True

    '''
    return assoc(d, key, value_function(d))

@curry
def update_key_v(key: Hashable, value_function, d, default=None):
    '''Update key's value for a given dictionary. Will add key if it
    doesn't exist.

    Args:
      key (Hashable): key to be updated
    
      value_function (Callable[[Any], Any]): function that takes the
        current value of key (or default) and returns a value for the
        key

      default (Any=None): default value to be provided to the
        value_function if the key doesn't already exist in the
        dictionary

      d (dict): dictionary to transform (with no side effects)

    Returns: (dict) new state of dictionary

    Examples:

    >>> new = pipe({'b': 2}, update_key_v('a', lambda v: v + 5, default=0))
    >>> new == {'a': 5, 'b': 2}
    True
    >>> new = pipe({'a': 4}, update_key_v('a', lambda v: v + 5, default=0))
    >>> new == {'a': 9}
    True

    '''
    return assoc(d, key, value_function(d.get(key, default)))

@curry
def only_if_key(key, func, d):
    '''Return func(d) if key in d, otherwise return d

    '''
    return func(d) if key in d else d

@curry
def update_if_key_exists(key: Hashable, value_function, d):
    '''Update key only if it already exists

    Args:
      key (Hashable): key to be updated (if the key already exists)
    
      value_function (Callable[[dict], Any]): function that takes the
        dictionary and returns a value for the key

      d (dict): dictionary to transform (with no side effects)

    Returns: (dict) new state of dictionary

    Examples:

    >>> pipe({}, update_if_key_exists('a', lambda d: d['a'] + 5))
    {}
    >>> pipe({'a': 4}, update_if_key_exists('a', lambda d: d['a'] + 5))
    {'a': 9}

    '''
    if key in d:
        return assoc(d, key, value_function(d))
    return d

@curry
def set_key(key: Hashable, value, d):
    '''Curried assoc

    Args:
      key (Hashable): key to be updated
    
      value (Any): value for the key

      d (dict): dictionary to transform (with no side effects)

    Returns: (dict) new state of dictionary

    Examples:

    >>> new = pipe({'b': 2}, set_key('a', 5))
    >>> new == {'a': 5, 'b': 2}
    True

    '''
    return assoc(d, key, value)

@curry
def drop_key(key: Hashable, d):
    ''' Curried dissoc

    Args:
      key (Hashable): key to be removed
    
      d (dict): dictionary to transform (with no side effects)

    Returns: (dict) new state of dictionary

    Examples:

    >>> pipe({'b': 2}, drop_key('b'))
    {}
    >>> pipe({'a': 2}, drop_key('b'))
    {'a': 2}
    
    '''
    return dissoc(d, key)
remove_key = drop_key

@curry
def drop_keys(keys: Iterable[Hashable], d):
    '''Curried dissoc (multiple keys)

    Args:
      keys (Iterable[Hashable]): keys to be removed
    
      d (dict): dictionary to transform (with no side effects)

    Returns: (dict) new state of dictionary

    Examples:

    >>> pipe({'a': 1, 'b': 2}, drop_keys(['a', 'b']))
    {}
    >>> pipe({'a': 2, 'b': 2}, drop_keys(['b', 'c']))
    {'a': 2}

    '''
    return dissoc(d, *keys)
remove_keys = drop_keys

@curry
def merge_keys(from_: Iterable[Hashable], to: Hashable, value_function, d):
    '''Merge multiple keys into a single key

    Args:
      from_ (Iterable[Hashable]): keys to be merged

      to (Hashable): key into which the from_ will be merged

      value_function (Callable[[dict], Any]): function that takes the
        dictionary and returns a value for the key given by "to"
        parameter

      d (dict): dictionary to transform (with no side effects)

    Returns: (dict) new state of dictionary

    Examples:

    >>> pipe(
    ...   {'a': 1, 'b': 2},
    ...   merge_keys(['a', 'b'], 'c', lambda d: d['a'] + d['b']),
    ... )
    {'c': 3}

    '''
    value = value_function(d)
    return pipe(d, drop_keys(from_), set_key(to, value))

@curry
def replace_key(k1, k2, value_function, d):
    '''Drop key k1 and replace with k2 if k1 exists

    Args:
      k1 (Hashable): key to drop

      k2 (Hashable): key to replace k1

      value_function (Callable[[dict], Any]): function that takes the
        dictionary and returns a value for the k2 key

      d (dict): dictionary to transform (with no side effects)

    Returns: (dict) new state of dictionary

    Examples:

    >>> pipe(
    ...   {'a': 1},
    ...   replace_key('a', 'c', lambda d: d['a'] + 2),
    ... )
    {'c': 3}

    '''
    if k1 in d:
        return merge_keys([k1], k2, value_function or getitem(k1), d)
    return d
switch_keys = replace_key
# @curry
# def switch_keys(k1, k2, value_function, d):
#     return pipe(
#         assoc(d, k2, value_function(d)),
#         drop_key(k1)
#     )

@curry
def valmaprec(func, d, **kw):
    '''Recursively map values of a dictionary (traverses Mapping and
    Sequence objects) using a function

    '''
    if is_dict(d):
        return pipe(
            d.items(),
            vmap(lambda k, v: (k, valmaprec(func, v, **kw))),
            type(d),
        )
    elif is_seq(d):
        return pipe(
            d, map(valmaprec(func, **kw)), type(d),
        )
    else:
        return func(d)

@curry
def match_d(match: dict, d: dict, *, default=Null):
    '''Given a match dictionary {key: regex}, return merged groupdicts
    only if all regexes are in the values (i.e. via regex.search) for
    all keys. Otherwise return default=Null.

    Args:

      match (Dict[Hashable, str]): Mapping of keys (in `d`) to regexes
        (as strings), where the regexes have named groups
        (e.g. r'(?P<group_name>regex)') somewhere in them

      d (dict): dictionary whose values (for keys given in `match`)
        must match (via search) the given regexes for these keys

      default (Any = Null): default value returned if either the
        dictionary d does not contain all the keys in `match` or not
        all of the regexes match

    Examples:

    >>> matched = pipe(
    ...   {'a': 'hello', 'b': 'world'},
    ...   match_d({'a': r'h(?P<v0>.*)', 'b': r'(?P<v1>.*)d'}),
    ... )
    >>> matched == {'v0': 'ello', 'v1': 'worl'}
    True
    >>> o = pipe(                   # missing a key
    ...   {'a': 'hello'},
    ...   match_d({'a': r'h(?P<v0>.*)', 'b': r'w(?P<v1>.*)'}),
    ... )
    >>> str(o)
    'Null'
    >>> matched = pipe(         # regexes don't match
    ...   {'a': 'hello', 'b': 'world'},
    ...   match_d({'a': r'ckjv(?P<v0>.*)', 'b': r'dkslfjl(?P<v1>.*)'}),
    ... )
    >>> str(matched)
    'Null'

    '''
    if set(d).issuperset(set(match)):
        if all(re.search(match[k], d[k]) for k in match):
            return merge(*(
                re.search(match[k], d[k]).groupdict()
                for k in match
            ))
    return default

@curry
def bakedict(key_f, value_f, iterable):
    new = {}
    for v in iterable:
        key = key_f(v)
        values = new.setdefault(key, [])
        values.append(value_f(v))
    return new

@curry
def vbakedict(key_f, value_f, iterable):
    return bakedict(vcall(key_f), vcall(value_f), iterable)


# ----------------------------------------------------------------------
#
# IP address functions
#
# ----------------------------------------------------------------------

def current_ip(ip_version):
    '''Returns the IP address (for a given version) of the interface where
    the default gateway is found

    '''
    return maybe_pipe(
        netifaces.gateways(),
        getitem('default'),
        getitem(ip_version),
        second,
        netifaces.ifaddresses,
        getitem(ip_version),
        maybe_first,
        lambda d: ip_interface(
            f'{d["addr"]}/{get_slash_from_mask(d["netmask"])}'
        )
    )

current_ipv4 = partial(current_ip, netifaces.AF_INET)
current_ipv6 = partial(current_ip, netifaces.AF_INET6)

def is_ipv4(ip: (str, int)):
    try:
        return ip_address(ip).version == 4
    except ValueError:
        return False

def is_ip(ip: (str, int)):
    try:
        ip_address(ip)
        return True
    except ValueError:
        return False

def is_interface(iface):
    try:
        ip_interface(iface)
        return True
    except ValueError:
        return False

def is_network(inet):
    try:
        ip_network(inet)
        return True
    except ValueError:
        return False

def get_slash(inet: Union[str, ip_network]):
    return 32 - int(math.log2(ip_network(inet).num_addresses))

def get_slash_from_mask(mask: str):
    addr = ip_interface(mask).ip
    max_slash = 32 if addr.version == 4 else 128
    max_int = 2**32 if addr.version == 4 else 2**128
    return max_slash - int(math.log2(max_int - int(addr)))

def is_comma_sep_ip(cs_ip):
    return ',' in cs_ip and all(is_ip(v) for v in cs_ip.split(','))

def is_ip_range(ip_range):
    if '-' in ip_range:
        parts = ip_range.split('-')
        if len(parts) == 2:
            base, last = parts
            if is_ipv4(base) and last.isdigit() and (0 <= int(last) <= 255):
                return True
    return False

def ip_to_seq(ip):
    if is_ip(ip):
        return [ip]
    elif is_network(ip):
        return pipe(ip_network(ip).hosts(), map(str), tuple)
    elif is_interface(ip):
        return pipe(ip_interface(ip).network.hosts(), map(str), tuple)
    elif is_comma_sep_ip(ip):
        return ip.split(',')
    elif is_ip_range(ip):
        base, last = ip.split('-')
        base = ip_address(base)
        last = int(last)
        first = int(str(base).split('.')[-1])
        return [str(ip_address(int(base) + i))
                for i in range(last - first + 1)]
    else:
        log.error(f'Unknown/unparsable ip value: {ip}')
        return []

def ip_tuple(ip):
    return pipe(str(ip).split('.'), map(int), tuple)

def sortips(ips):
    return sort_by(compose(ip_address, strip, strip_comments), ips)
sort_ips = sortips

def get_ips_from_file(path):
    return get_ips_from_content(Path(path).read_text())

def get_ips_from_content(content):
    return get_ips_from_lines(content.splitlines())
get_ips_from_str = get_ips_from_content

def get_ips_from_lines(lines):
    return pipe(
        lines,
        map(to_str),
        strip_comments,
        filter(strip),
        mapcat(ip_re.findall),
        # filter(is_ip),
        # mapcat(ip_to_seq),
        tuple,
    )

def get_networks_from_file(path):
    return get_networks_from_content(Path(path).expanduser().read_text())

def get_networks_from_content(content):
    return get_networks_from_lines(content.splitlines())

def get_networks_from_lines(lines):
    return pipe(
        lines,
        map(to_str),
        strip_comments,
        filter(strip),
        filter(is_network),
        tuple,
    )

@curry
def in_ip_range(ip0, ip1, ip):
    start = int(ip_address(ip0))
    stop = int(ip_address(ip1))
    return int(ip_address(ip)) in range(start, stop + 1)

def zpad(ip):
    '''Zero-pad an IP address

    Examples:
    
    >>> zpad('1.2.3.4')
    '001.002.003.004'

    '''
    return '.'.join(s.zfill(3) for s in str(ip).strip().split('.'))

def unzpad(ip):
    '''Remove zero-padding from an IP address

    Examples:
    
    >>> unzpad('001.002.003.004')
    '1.2.3.4'

    '''
    return pipe(ip.split('.'), map(int), map(str), '.'.join)


# ----------------------------------------------------------------------
#
# Basic type operations
#
# ----------------------------------------------------------------------

def is_str(v):
    return isinstance(v, str)
is_not_string = complement(is_str)

def to_str(content, encoding='utf-8', errors='ignore'):
    if type(content) is bytes:
        return content.decode(encoding, errors)
    else:
        return str(content)

def to_bytes(content, encoding='utf-8', errors='ignore'):
    if type(content) is str:
        return content.encode(encoding, errors)
    elif type(content) is bytes:
        return content
    else:
        return str(content).encode(encoding, errors)

def is_dict(d):
    return isinstance(d, collections.abc.Mapping)
is_not_dict = complement(is_dict)

def is_indexable(s):
    return hasattr(s, '__getitem__')

def is_seq(s):
    return (isinstance(s, collections.abc.Iterable) and not
            is_dict(s) and not
            isinstance(s, (str, bytes)))
is_not_seq = complement(is_seq)

def flatdict(obj: Union[dict, Any], keys=()):
    '''Flatten a Python dictionary such that nested values are returned
    with the key sequence required to access them.

    Examples:

    >>> pipe({'a': {'b': [1, 2, 3]}, 'c': 2}, flatdict, list)
    [('a', 'b', [1, 2, 3]), ('c', 2)]
    '''
    if is_dict(obj):
        for k, v in obj.items():
            yield from flatdict(v, keys + (k, ))
    else:
        yield keys + (obj,)


# ----------------------------------------------------------------------
#
# Time-oriented functions
#
# ----------------------------------------------------------------------

def ctime(path: Union[str, Path]):
    return Path(path).stat().st_ctime

def maybe_dt(ts):
    '''Parse ts to datetime object (using dateutil.parser.parse) or return
    Null

    '''
    try:
        return dateutil.parser.parse(ts)
    except ValueError:
        return Null

def parse_dt(ts: str, local=False):
    dt = dateutil.parser.parse(ts)
    if local:
        return dt.astimezone(dateutil.tz.tzlocal())
    return dt

def ctime_as_dt(path: Union[str, Path]):
    return pipe(
        path,
        ctime,
        datetime.fromtimestamp,
    )
dt_ctime = ctime_as_dt

def to_dt(value, default=datetime.fromtimestamp(0)):
    '''Attempt to parse the given value as a datetime object, otherwise
    return default=epoch

    Will try:
    - dateutil.parser.parse
    - 20190131T130506123456 (i.e. with microseconds)

    '''
    try_except = [
        (lambda v: dateutil.parser.parse(v), (ValueError, TypeError)),
        (lambda v: datetime.strptime(v, "%Y%m%dT%H%M%S%f"),
         (ValueError, TypeError)),
    ]
    for func, excepts in try_except:
        try:
            output = func(value)
            return output
        except excepts:
            continue
    return default


# ----------------------------------------------------------------------
#
# Import functions
#
# ----------------------------------------------------------------------

def function_from_path(func_path: str):
    '''Return the function object for a given module path

    '''
    return pipe(
        func_path,
        lambda path: path.rsplit('.', 1),
        vcall(lambda mod_path, func_name: (
            importlib.import_module(mod_path), func_name
        )),
        vcall(lambda mod, name: (
            (name, _getattr(mod, name))
            if hasattr(mod, name) else
            (name, None)
        )),
    )


def strip(content):
    return content.strip()

@dispatch(str)
def strip_comments(line):
    return line[:line.index('#')] if '#' in line else line
    
@dispatch((collections.abc.Iterable, list, tuple, PVector))  # noqa
def strip_comments(lines):
    return pipe(
        lines,
        map(strip_comments),
        tuple,
    )

def remove_comments(lines):
    return pipe(
        lines,
        filter(lambda l: not l.startswith('#')),
    )

def help_text(s):
    return textwrap.shorten(s, 1e300)

def clipboard_copy(content):
    import pyperclip
    pyperclip.copy(content)

def clipboard_paste():
    import pyperclip
    return pyperclip.paste()

def xlsx_to_clipboard(content):
    return pipe(
        content,
        to_str,
        lambda c: c if c.endswith('\n') else c + '\n',
        clipboard_copy,
    )

def escape_row(row):
    return pipe(
        row,
        map(lambda v: v.replace('"', '""')),
        '\t'.join,
    )

def output_rows_to_clipboard(rows):
    return pipe(
        rows,
        map(escape_row),
        '\n'.join,
        clipboard_copy,
    )

def difflines(A, B):
    linesA = pipe(
        A.splitlines(),
        strip_comments,
        filter(None),
        set,
    )
    linesB = pipe(
        B.splitlines(),
        strip_comments,
        filter(None),
        set,
    )
    return pipe(linesA - linesB, sorted)

def intlines(A, B):
    linesA = pipe(
        A.splitlines(),
        strip_comments,
        filter(None),
        set,
    )
    linesB = pipe(
        B.splitlines(),
        strip_comments,
        filter(None),
        set,
    )
    return pipe(linesA & linesB, sorted)

@curry
def peek(nbytes, path):
    with Path(path).open('r', encoding='latin-1') as rfp:
        return rfp.read(nbytes)

def backup_path(path):
    path = Path(path)
    dt = dt_ctime(path)
    return Path(
        path.parent,
        ''.join((
            'backup',
            '-',
            path.stem,
            '-',
            dt.strftime('%Y-%m-%d_%H%M%S'),
            path.suffix
        ))
    )
    
def arg_intersection(func, kw):
    params = inspect.signature(func).parameters
    if any(p.kind == p.VAR_KEYWORD for p in params.values()):
        return kw
    else:
        return {k: kw[k] for k in set(params) & set(kw)}

def positional_only_args(func):
    return pipe(
        inspect.signature(func).parameters.values(),
        filter(
            lambda p: p.kind not in {p.VAR_KEYWORD,
                                     p.KEYWORD_ONLY,
                                     p.VAR_POSITIONAL}
        ),
        filter(lambda p: p.default == p.empty),
        map(lambda p: p.name),
        tuple,
    )

def is_arg_superset(kwargs, func):
    '''Does the kwargs dictionary contain the func's required params?

    '''
    return pipe(
        func,
        positional_only_args,
        set(kwargs).issuperset,
    )

@curry
def regex_transform(regexes, text):
    '''Given a sequence of [(regex, replacement_text), ...] pairs,
    transform text by making all replacements

    '''
    if not is_str(text):
        return text

    regexes = pipe(
        regexes,
        vmap(lambda regex, replace: (re.compile(regex), replace)),
        tuple,
    )
    for regex, replace in regexes:
        text = regex.sub(replace, text)
    return text

@curry
def from_edgelist(edgelist, factory=None):
    '''Curried nx.from_edgelist

    '''
    import networkx as nx
    return nx.from_edgelist(edgelist, create_using=factory)

@curry
def bfs_tree(G, source, reverse=False, depth_limit=None):
    '''Curried nx.tranversal.bfs_tree

    '''
    import networkx as nx
    return nx.traversal.bfs_tree(
        G, source, reverse=reverse,
        depth_limit=depth_limit
    )

@curry
def contains(value, obj):
    '''Curried in operator

    '''
    return value in obj




import logging
from pathlib import Path
from functools import wraps
from typing import Any
from collections.abc import Mapping, Iterable

import ruamel.yaml
from ruamel.yaml.comments import CommentedMap
from toolz import pipe
from multipledispatch import dispatch

from .common import no_pyrsistent

log = logging.getLogger('yaml')
log.addHandler(logging.NullHandler())

@wraps(ruamel.yaml.dump)
def dump(*a, **kw) -> str:
    kw['Dumper'] = ruamel.yaml.RoundTripDumper
    kw['default_flow_style'] = False
    kw['width'] = 2**31
    return ruamel.yaml.dump(*a, **kw)

@wraps(ruamel.yaml.load)
def load(*a, **kw) -> Any:
    kw['Loader'] = ruamel.yaml.RoundTripLoader
    return ruamel.yaml.load(*a, **kw)
    
def read_yaml(path: (str, Path)):
    '''Read YAML data from path and return object
    '''
    with Path(path).expanduser().open() as rfp:
        return load(rfp)

def _write_yaml(path, data):
    with Path(path).open('w') as wfp:
        dump(data, wfp)
    return True

@dispatch((str, Path), Mapping)
def write_yaml(path, dict_data):
    '''Write data as YAML to path

    Args:
      path (str, Path): path to write to

      data (dict-like): dictionary-like object to write (will be
         recursively converted to base Python types)

    Returns: (bool) success of write operation,

    Raises: on error, will raise exception

    '''
    return _write_yaml(path, pipe(dict_data, no_pyrsistent, CommentedMap))

@dispatch((str, Path), Iterable)  # noqa
def write_yaml(path, iterable_data):
    '''Write data as YAML to path

    Args:
      path (str, Path): path to write to

      data (sequence): sequence object to write (will be recursively
         converted to base Python types)

    Returns: (bool) success of write operation,

    Raises: on error, will raise exception

    '''
    return _write_yaml(path, pipe(iterable_data, no_pyrsistent))

@dispatch((str, Path), object)  # noqa
def write_yaml(path, object_data):
    return _write_yaml(path, object_data)




import sys
from pathlib import Path
import logging
from typing import Union

import click
from toolz.curried import curry

from ..common import clipboard_paste

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

@curry
def exit_with_msg(logger, msg):
    ctx = click.get_current_context()
    click.echo(ctx.get_help())
    logger.error(msg)
    raise click.Abort()

_exit_with_msg = exit_with_msg(log)

def get_input_content(inpath, clipboard=False):
    if inpath:
        content = Path(inpath).read_text()
    elif clipboard:
        content = clipboard_paste()
    else:
        content = sys.stdin.read()
    return content

def path_cb_or_stdin(inpath: Union[str, Path], clipboard: bool):
    if inpath:
        log.info(f'Getting input from path: {inpath}')
        return Path(inpath).read_text()
    elif clipboard:
        log.info('Getting input from clipboard...')
        return clipboard_paste()
    log.info('Getting input from stdin...')
    return sys.stdin.read()

def cb_or_stdin(clipboard):
    if clipboard:
        log.info('Getting input from clipboard...')
        return clipboard_paste()
    log.info('Getting input from stdin...')
    return sys.stdin.read()



import os
import sys
import subprocess
import shlex
import logging
from threading import Timer
from typing import Callable, Any
from pathlib import Path        # noqa: for doctest
import tempfile                 # noqa: for doctest

from toolz.curried import merge, map, pipe, curry
from toolz.curried import do

from .common import cprint

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

@curry
def shell_iter(command, *, echo: bool = True,
               echo_func: Callable[[Any], None] = cprint(file=sys.stderr,
                                                         end=''),
               timeout: int = None, **popen_kw):
    '''Execute a shell command, yield lines of output as they come
    possibly echoing command output to a given echo_func, and finally
    yields the status code of the process.

    This will run the shell command, yielding each line of output as
    it runs. When the process terminates, it will then yield the
    remainer of output, then finally the integer status code. It can
    also be terminated early via a timeout parameter. By default, the
    command will also echo to stderr.

    Args:

      command (str): Shell command to execute. Tilde (~) and shell
        variable completion provided

      echo (bool): Should the output be echoed to echo_func in
        addition to yielding lines of output?

      echo_func (Callable[[Any], None]): Function to use when echoing
        output. **Be warned**, this funciton is called __for each
        character__ of output. By default, this is `cprint(end='')`
        (i.e. print with end='')

      timeout (int): If set, the process will be killed after this
        many seconds (kill -9).

    Returns: generator of the form

        *output_lines, status_code = shell_iter(...)

      where output_lines is a sequence of strings of output and
      status_code is an integer status code

    Examples:

    >>> with tempfile.TemporaryDirectory() as tempdir:
    ...     root = Path(tempdir)
    ...     _ = Path(root, 'a.txt').write_text('')
    ...     _ = Path(root, 'b.txt').write_text('')
    ...     *lines, status = shell_iter(f'ls {root}')
    a.txt
    b.txt
    >>> lines
    ['a.txt', 'b.txt']
    >>> status
    0

    >>> with tempfile.TemporaryDirectory() as tempdir:
    ...     root = Path(tempdir)
    ...     _ = Path(root, 'c.txt').write_text('')
    ...     _ = Path(root, 'd.txt').write_text('')
    ...     *lines, _ = shell_iter(f'ls {root}', echo=False)
    >>> lines == ['c.txt', 'd.txt']
    True

    >>> *lines, status = shell_iter(
    ...     f'sleep 5', echo=False, timeout=0.01
    ... )
    >>> lines
    []
    >>> status
    -9

    '''
    popen_kw = merge({
        'stdout': subprocess.PIPE,
        'stderr': subprocess.PIPE,
    }, popen_kw)

    command_split = pipe(
        shlex.split(command),
        map(os.path.expanduser),
        map(os.path.expandvars),
        tuple,
    )
    process = subprocess.Popen(command_split, **popen_kw)

    if timeout:
        # https://www.blog.pythonlibrary.org/2016/05/17/python-101-how-to-timeout-a-subprocess/
        def kill():
            log.warning(f'Process ({command_split[0]}) timeout expired.')
            return process.kill()
        timer = Timer(timeout, kill)
        timer.start()
    
    def process_running():
        return process.poll() is None

    line = ''
    while process_running():
        char = process.stdout.read(1).decode('utf-8', errors='ignore')
        if char:
            echo_func(char) if echo else ''
            if char == '\n':
                yield line
                line = ''
            else:
                line += char

    if timeout:
        timer.cancel()
    rest = process.stdout.read().decode('utf-8', errors='ignore')
    for char in rest:
        echo_func(char) if echo else ''
        if char == '\n':
            yield line
            line = ''
        else:
            line += char

    if line:
        echo_func(char) if echo else ''
        yield line

    yield process.poll()

@curry
def shell(command, **kw):
    '''Execute a shell command, yield lines of output as they come
    possibly echoing command output to a given echo_func, and finally
    yields the status code of the process.

    This will run the shell command, yielding each line of
    output. When the process terminates, it will then yield the status
    code.

    Args:

      command (str): Shell command to execute. Tilde (~) and shell
        variable completion provided

      echo (bool): Should the output be echoed to echo_func in
        addition to yielding lines of output?

      echo_func (Callable[[Any], None]): Function to use when echoing
        output. **Be warned**, this funciton is called __for each
        character__ of output. By default, this is `cprint(end='')`
        (i.e. print with end='')

      timeout (int): If set, the process will be killed after this
        many seconds (kill -9).

    Examples:

    >>> with tempfile.TemporaryDirectory() as tempdir:
    ...     root = Path(tempdir)
    ...     _ = Path(root, 'a.txt').write_text('')
    ...     _ = Path(root, 'b.txt').write_text('')
    ...     *lines, status = shell_iter(f'ls {root}')
    a.txt
    b.txt
    >>> lines
    ['a.txt', 'b.txt']
    >>> status
    0

    >>> with tempfile.TemporaryDirectory() as tempdir:
    ...     root = Path(tempdir)
    ...     _ = Path(root, 'c.txt').write_text('')
    ...     _ = Path(root, 'd.txt').write_text('')
    ...     *lines, _ = shell_iter(f'ls {root}', echo=False)
    >>> lines == ['c.txt', 'd.txt']
    True

    >>> *lines, status = shell_iter(
    ...     f'sleep 5', echo=False, timeout=0.01
    ... )
    >>> lines
    []
    >>> status
    -9
    '''
    *lines, status = shell_iter(command, **kw)
    return status, '\n'.join(lines)

@curry
def getoutput(command, **kw):
    status, content = shell(command, **kw)
    return content



import re
from pathlib import Path

from setuptools import setup, find_packages

HERE = Path(__file__).resolve().parent

version_re = re.compile(r"^__version__\s*=\s*'(?P<version>.*)'$", re.M)
def version():
    match = version_re.search(Path('larc/__init__.py').read_text())
    if match:
        return match.groupdict()['version'].strip()
    return '0.0.1'

long_description = Path(HERE, 'README.md').resolve().read_text()

setup(
    name='larc',
    packages=find_packages(
        exclude=['tests'],
    ),
    package_dir={
        'larc': 'larc',
    },

    install_requires=[
        'toolz',
        'multipledispatch',
        'pyrsistent',
        'networkx',
        'coloredlogs',
        'gevent',
        'ruamel.yaml',
        'ipython',
        'click',
        'pyperclip',
        'python-nmap',
        'tcping',
        'ipparser',
        'dnspython',
        'paramiko',
        'xmljson',
        'python-dateutil',
        'jmespath',
        'netifaces',
    ],

    version=version(),
    description=('Collection of helper functions and general utilities'
                 ' used across various LARC projects'),
    long_description=long_description,
    long_description_content_type='text/markdown',

    url='https://github.org/lowlandresearch/larc',

    author='Lowland Applied Research Company (LARC)',
    author_email='dogwynn@lowlandresearch.com',

    license='MIT',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
    ],

    zip_safe=False,

    keywords=('utilities functional toolz'),

    scripts=[
    ],

    entry_points={
        'console_scripts': [
            'diffips=larc.cli.ips:diff_ips',
            'intips=larc.cli.ips:int_ips',
            'difflines=larc.cli.text:diff_lines',
            'intlines=larc.cli.text:int_lines',
            'sortips=larc.cli.ips:sort_ips',
            'getips=larc.cli.ips:get_ips',
            'getsubnets=larc.cli.ips:get_subnets',
            'zpad=larc.cli.ips:zpad_ips',
            'unzpad=larc.cli.ips:unzpad_ips',
        ],
    },
)
